/*
 * (C) Copyright 2012 Michal Simek <monstr@monstr.eu>
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

#include <common.h>
#include <fdtdec.h>
#include <fpga.h>
#include <mmc.h>
#include <netdev.h>
#include <zynqpl.h>
#if defined(CONFIG_TARGET_ZYNQ_UCB7020)
#include <status_led.h>
#endif
#include <asm/arch/hardware.h>
#include <asm/arch/sys_proto.h>

#include <asm/arch-zynq/bootconfig.h>
#include <status_led.h>
#include <i2c.h>

/* storage for system vaiables */
static char bootstat_var[6];

static int isValidOption(uint8_t boot_option);

DECLARE_GLOBAL_DATA_PTR;

int ucb7020_bootopt_found = 0;

#if (defined(CONFIG_FPGA) && !defined(CONFIG_SPL_BUILD)) || \
    (defined(CONFIG_SPL_FPGA_SUPPORT) && defined(CONFIG_SPL_BUILD))
static xilinx_desc fpga;

/* It can be done differently */
static xilinx_desc fpga010 = XILINX_XC7Z010_DESC(0x10);
static xilinx_desc fpga015 = XILINX_XC7Z015_DESC(0x15);
static xilinx_desc fpga020 = XILINX_XC7Z020_DESC(0x20);
static xilinx_desc fpga030 = XILINX_XC7Z030_DESC(0x30);
static xilinx_desc fpga035 = XILINX_XC7Z035_DESC(0x35);
static xilinx_desc fpga045 = XILINX_XC7Z045_DESC(0x45);
static xilinx_desc fpga100 = XILINX_XC7Z100_DESC(0x100);
#endif

int board_init(void)
{
#if defined(CONFIG_ENV_IS_IN_EEPROM) && !defined(CONFIG_SPL_BUILD) && \
    defined(CONFIG_ZYNQ_EVALBOARD)
	unsigned char eepromsel = CONFIG_SYS_I2C_MUX_EEPROM_SEL;
#endif
#if (defined(CONFIG_FPGA) && !defined(CONFIG_SPL_BUILD)) || \
    (defined(CONFIG_SPL_FPGA_SUPPORT) && defined(CONFIG_SPL_BUILD))
	u32 idcode;

#if defined(CONFIG_TARGET_ZYNQ_UCB7020) && !defined(CONFIG_SPL_BUILD)
	yellow_led_on();	/* set yellow led ON */
	red_led_off();	/* set red led OFF */
#endif
	idcode = zynq_slcr_get_idcode();

	switch (idcode) {
	case XILINX_ZYNQ_7010:
		fpga = fpga010;
		break;
	case XILINX_ZYNQ_7015:
		fpga = fpga015;
		break;
	case XILINX_ZYNQ_7020:
		fpga = fpga020;
		break;
	case XILINX_ZYNQ_7030:
		fpga = fpga030;
		break;
	case XILINX_ZYNQ_7035:
		fpga = fpga035;
		break;
	case XILINX_ZYNQ_7045:
		fpga = fpga045;
		break;
	case XILINX_ZYNQ_7100:
		fpga = fpga100;
		break;
    default:
        return -1;
	}
#endif

#if (defined(CONFIG_FPGA) && !defined(CONFIG_SPL_BUILD)) || \
    (defined(CONFIG_SPL_FPGA_SUPPORT) && defined(CONFIG_SPL_BUILD))
	fpga_init();
	(void)fpga_add(fpga_xilinx, &fpga);
#endif
#if defined(CONFIG_ENV_IS_IN_EEPROM) && !defined(CONFIG_SPL_BUILD) && \
    defined(CONFIG_ZYNQ_EVALBOARD)
	if (eeprom_write(CONFIG_SYS_I2C_MUX_ADDR, 0, &eepromsel, 1))
		puts("I2C:EEPROM selection failed\n");
#endif
	return 0;
}

extern int ramdisk_init(unsigned int start_addr, unsigned len);

int board_late_init(void)
{
	u8 hw_boot_reason, boot_option;
    unsigned long addr;
    unsigned long len;
    int i, sum, sum2;
    int ret;
    uint8_t ethaddr[6];     /* ethernet address */
    uint8_t serialNum[8];    /* serial number */
    char env_var[40];
    uint8_t chkSum;         /* simple checksum */

	/* printf("%s %d POSTFAIL set to 0x%x\n", __FUNCTION__, __LINE__, gd->flags); */
	if (gd->flags & GD_FLG_POSTFAIL) {
		setenv ("bootcmd", "");
		red_led_on();	/* turn off yellow */
		yellow_led_off();	/* turn ON red */
		printf("Boot process aborted due to POST errors\n");
		return 1;
	}

	switch ((zynq_slcr_get_boot_mode()) & ZYNQ_BM_MASK) {
	case ZYNQ_BM_QSPI:
		/* setenv("modeboot", "qspiboot"); */
		setenv("modeboot", "new_boot");
		break;
	case ZYNQ_BM_NAND:
		setenv("modeboot", "nandboot");
		break;
	case ZYNQ_BM_NOR:
		setenv("modeboot", "norboot");
		break;
	case ZYNQ_BM_SD:
		setenv("modeboot", "sdboot");
		break;
	case ZYNQ_BM_JTAG:
		setenv("modeboot", "jtagboot");
		break;
	default:
		setenv("modeboot", "");
		break;
	}
/*
 *	get hardware boot reason and store it environment variable
 */
	hw_boot_reason = (uint8_t)zynq_get_boot_reason();
	strcpy (bootstat_var, simple_itoa(hw_boot_reason));
	setenv("bootstatus", bootstat_var);

/*
 *  get bootoption from EEPROM
 */
    ret = i2c_read(CONFIG_SYS_I2C_EEPROM_ADDR, EEPROM_OFS_BOOTOPT, 1,
                   (uint8_t *)&boot_option, 1);
    if ((ret) || !isValidOption(boot_option)) {
        printf("No valid bootoption, using %d instead\n", DEFAULT_BOOT_OPTION);
        boot_option = DEFAULT_BOOT_OPTION;
    }
    boot_option &= 0x7f;    /* cut off parity */
	strcpy (bootstat_var, simple_itoa(boot_option));
	setenv("bootoption", bootstat_var);

    ucb7020_bootopt_found = 1;  /* starting here, the bootoption is valid */

	if ((boot_option > 0) && (boot_option < 11)) {
		if (boot_option > 6) {
		    if (boot_option < 9) {  /* 7 & 8 -> A,B */
		    	setenv("pri_bank", BANK_A_BOOT_OFFSET_STR);
    			setenv("sec_bank", BANK_B_BOOT_OFFSET_STR);
    		}
    		else {  /* 9 & 10 -> B,A */
    			setenv("pri_bank", BANK_B_BOOT_OFFSET_STR);
    			setenv("sec_bank", BANK_A_BOOT_OFFSET_STR);
    		}
		}
		else if (boot_option & 1) {
			setenv("pri_bank", BANK_A_BOOT_OFFSET_STR);
			setenv("sec_bank", BANK_B_BOOT_OFFSET_STR);
		}
		else {
			setenv("pri_bank", BANK_B_BOOT_OFFSET_STR);
			setenv("sec_bank", BANK_A_BOOT_OFFSET_STR);
		}
	}
    /* strcpy (bootres_var, simple_itoa(boot_result));
     * setenv("bootresult", bootres_var);
     */
    addr = simple_strtoul(getenv("image_addr"), NULL, 16);
    len = simple_strtoul(getenv("image_len"), NULL, 16);
	(void)ramdisk_init(addr, len);

    /* read serial number and ethernet address from EEPROM if avilable */
    ret = i2c_read(CONFIG_SYS_I2C_EEPROM_ADDR, EEPROM_OFS_ETHADDR, 1,
                   (uint8_t *)ethaddr, 6);
    if (!ret) {
        ret = i2c_read(CONFIG_SYS_I2C_EEPROM_ADDR, EEPROM_OFS_SERNUM, 1,
                       (uint8_t *)serialNum, 8);
        if (!ret) {
            ret = i2c_read(CONFIG_SYS_I2C_EEPROM_ADDR, EEPROM_OFS_CHKSUM, 1,
                           (uint8_t *)&chkSum, 1);
        }
    }
    if (ret) {
        printf("%s %d i2c read returned %d\n", __FUNCTION__, __LINE__, ret);
        return ret;
    }
    /* test, if values are ok */
    sum = 0;
    /*lint -e{644, 574} */
    for (i = 0; i<sizeof(ethaddr); i++)
        sum += ethaddr[i];
    if (sum == 0) {
        printf("%s %d invalid ethaddr encountered in i2c\n", __FUNCTION__, __LINE__);
        goto done;
    }
    sum2 = 0;
    /*lint -e{644, 574} */
    for (i = 0; i<sizeof(serialNum); i++)
        sum2 += serialNum[i]; /*lint -e(644) */
    if (sum2 == 0) {
        printf("%s %d invalid serial number encountered in i2c\n", __FUNCTION__, __LINE__);
        goto done;
    }
    /*lint -e{644} */
    if ((uint8_t)((sum + sum2) & 0xff) != (uint8_t)~chkSum) {
        printf("%s %d invalid checksum encountered in i2c 0x%x exp: 0x%x\n",
            __FUNCTION__, __LINE__, chkSum, ~((sum + sum2) & 0xff)); /*lint -e(644) */
        goto done;
    }
    sprintf(env_var, "%02X:%02X:%02X:%02X:%02X:%02X", ethaddr[0], ethaddr[1], ethaddr[2],
        ethaddr[3], ethaddr[4], ethaddr[5]);
    setenv("ethaddr", env_var);
    /*lint -e{574} */
    for (i = 0; i < sizeof(serialNum); i++)
        env_var[i] = (char)serialNum[i];
    env_var[i] = '\0';
    setenv("serial#", env_var);
done:

	return 0;
}

#ifdef CONFIG_DISPLAY_BOARDINFO
int checkboard(void)
{
	puts("Board:\tXilinx Zynq\n");
	return 0;
}
#endif

int board_eth_init(bd_t *bis)
{
	int ret = 0;

#ifdef CONFIG_XILINX_AXIEMAC
	ret |= xilinx_axiemac_initialize(bis, XILINX_AXIEMAC_BASEADDR,
						XILINX_AXIDMA_BASEADDR);
#endif
#ifdef CONFIG_XILINX_EMACLITE
	u32 txpp = 0;
	u32 rxpp = 0;
# ifdef CONFIG_XILINX_EMACLITE_TX_PING_PONG
	txpp = 1;
# endif
# ifdef CONFIG_XILINX_EMACLITE_RX_PING_PONG
	rxpp = 1;
# endif
	ret |= xilinx_emaclite_initialize(bis, XILINX_EMACLITE_BASEADDR,
			txpp, rxpp);
#endif

#if defined(CONFIG_ZYNQ_GEM)
# if defined(CONFIG_ZYNQ_GEM0)
	ret |= zynq_gem_initialize(bis, ZYNQ_GEM_BASEADDR0,
				   CONFIG_ZYNQ_GEM_PHY_ADDR0,
				   CONFIG_ZYNQ_GEM_EMIO0);
# endif
# if defined(CONFIG_ZYNQ_GEM1)
	ret |= zynq_gem_initialize(bis, ZYNQ_GEM_BASEADDR1,
				   CONFIG_ZYNQ_GEM_PHY_ADDR1,
				   CONFIG_ZYNQ_GEM_EMIO1);
# endif
#endif
	return ret;
}

#ifdef CONFIG_CMD_MMC
/*lint -efunc(715,board_mmc_init) */
int board_mmc_init(bd_t *bd)
{
	int ret = 0;

#if defined(CONFIG_ZYNQ_SDHCI)
# if defined(CONFIG_ZYNQ_SDHCI0)
    /*lint -e{838} */
	ret = zynq_sdhci_init(ZYNQ_SDHCI_BASEADDR0);
# endif
# if defined(CONFIG_ZYNQ_SDHCI1)
	ret |= zynq_sdhci_init(ZYNQ_SDHCI_BASEADDR1);
# endif
#endif
	return ret;
}
#endif

/*
 *  return 1 if the parity bit in the MSB is correct
 *  return 0 otherwise
 */
static int isValidOption(uint8_t boot_option)
{
    int bParity;
    
    bParity = (boot_option & 0x80) >> 7;
    boot_option &= 0x7f;
    /* this could be done much smarter but there is an optimizing compiler
       that could do it as well... */
    switch (boot_option) {
        case 3:
        case 5:
        case 6:
        case 9:
        case 10:
            if (bParity == 0)
                return 1;
            else
                return 0;
        case 1:
        case 2:
        case 4:
        case 7:
        case 8:
            if (bParity == 0)
                return 0;
            else
                return 1;
        default:
            return 0;
    }
}

int dram_init(void)
{
#ifdef CONFIG_OF_CONTROL
	int node;
	fdt_addr_t addr;
	fdt_size_t size;
	const void *blob = gd->fdt_blob;

	node = fdt_node_offset_by_prop_value(blob, -1, "device_type",
					     "memory", 7);
	if (node == -FDT_ERR_NOTFOUND) {
		debug("ZYNQ DRAM: Can't get memory node\n");
		return -1;
	}
	addr = fdtdec_get_addr_size(blob, node, "reg", &size);
	if (addr == FDT_ADDR_T_NONE || size == 0) {
		debug("ZYNQ DRAM: Can't get base address or size\n");
		return -1;
	}
	gd->ram_size = size;
#else
	gd->ram_size = CONFIG_SYS_SDRAM_SIZE;
#endif
	zynq_ddrc_init();

	return 0;
}
