/*
 * Copyright (c) 2001 William L. Pitts
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are freely
 * permitted provided that the above copyright notice and this
 * paragraph and the following disclaimer are duplicated in all
 * such forms.
 *
 * This software is provided "AS IS" and without any express or
 * implied warranties, including, without limitation, the implied
 * warranties of merchantability and fitness for a particular
 * purpose.
 */

#include <common.h>
#include <net.h>
#include <bootm.h>
#include <command.h>
#include <env_callback.h>
#include <linux/ctype.h>
#include <net.h>
#include <elf.h>
#include <vxworks.h>
#ifdef CONFIG_TARGET_ZYNQ_UCB7020
#include <asm/io.h>
#include <i2c.h>
#include <asm/arch-zynq/bootconfig.h>
#include <asm/arch-zynq/hardware.h>
#include <spi.h>
#include <spi_flash.h>
#endif
/* #include <version.h> */

#if defined(CONFIG_WALNUT) || defined(CONFIG_SYS_VXWORKS_MAC_PTR)
DECLARE_GLOBAL_DATA_PTR;
#endif

static unsigned long load_elf_image_phdr(unsigned long addr);
static unsigned long load_elf_image_shdr(unsigned long addr);

/* Allow ports to override the default behavior */
static unsigned long do_bootelf_exec(ulong (*entry)(int, char * const[]),
			       int argc, char * const argv[])
{
	unsigned long ret;

	/*
	 * QNX images require the data cache is disabled.
	 * Data cache is already flushed, so just turn it off.
	 */
	int dcache = dcache_status();
	if (dcache)
		dcache_disable();

	/*
	 * pass address parameter as argv[0] (aka command name),
	 * and all remaining args
	 */
	ret = entry(argc, argv);

	if (dcache)
		dcache_enable();

	return ret;
}

/* ======================================================================
 * Determine if a valid ELF image exists at the given memory location.
 * First looks at the ELF header magic field, the makes sure that it is
 * executable and makes sure that it is for a PowerPC.
 * ====================================================================== */
int valid_elf_image(unsigned long addr)
{
	Elf32_Ehdr *ehdr;		/* Elf header structure pointer */

	/* -------------------------------------------------- */

	ehdr = (Elf32_Ehdr *) addr;

	if (!IS_ELF(*ehdr)) {
		printf("## No elf image at address 0x%08lx\n", addr);
		return 0;
	}

	if (ehdr->e_type != ET_EXEC) {
		printf("## Not a 32-bit elf image at address 0x%08lx\n", addr);
		return 0;
	}

#if 0
	if (ehdr->e_machine != EM_PPC) {
		printf("## Not a PowerPC elf image at address 0x%08lx\n", addr);
		return 0;
	}
#endif

	return 1;
}

/* ======================================================================
 * Interpreter command to boot an arbitrary ELF image from memory.
 * ====================================================================== */
int do_bootelf(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	unsigned long addr;		/* Address of the ELF image     */
	unsigned long rc;		/* Return value from user code  */
	char *sload, *saddr;
	const char *ep = getenv("autostart");

	/* -------------------------------------------------- */
	int rcode = 0;

	sload = saddr = NULL;
	if (argc == 3) {
		sload = argv[1];
		saddr = argv[2];
	} else if (argc == 2) {
		if (argv[1][0] == '-')
			sload = argv[1];
		else
			saddr = argv[1];
	}

	if (saddr)
		addr = simple_strtoul(saddr, NULL, 16);
	else
		addr = load_addr;

	if (!valid_elf_image(addr))
		return 1;

	if (sload && sload[1] == 'p')
		addr = load_elf_image_phdr(addr);
	else
		addr = load_elf_image_shdr(addr);

	if (ep && !strcmp(ep, "no"))
		return rcode;

	printf("## Starting application at 0x%08lx ...\n", addr);

	/*
	 * pass address parameter as argv[0] (aka command name),
	 * and all remaining args
	 */
	rc = do_bootelf_exec((void *)addr, argc - 1, argv + 1);
	if (rc != 0)
		rcode = 1;

	printf("## Application terminated, rc = 0x%lx\n", rc);
	return rcode;
}

#define BUILD_BUF_SIZE 255
/* ======================================================================
 * Interpreter command to boot VxWorks from a memory image.  The image can
 * be either an ELF image or a raw binary.  Will attempt to setup the
 * bootline and other parameters correctly.
 * ====================================================================== */
/*lint -e{438, 550} for unused ipAddr */
int do_bootvx(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	unsigned long addr;		/* Address of image            */
	unsigned long bootaddr;	/* Address to put the bootline */
	char *bootline;			/* Text of the bootline        */
	char *tmp;			/* Temporary char pointer      */
	char build_buf[BUILD_BUF_SIZE];		/* Buffer for building the bootline */
    char *s;
	unsigned long ipAddr;		/* temp storage */
    char envTmp[22];
    unsigned int iTmp;

	/* ---------------------------------------------------
	 *
	 * Check the loadaddr variable.
	 * If we don't know where the image is then we're done.
	 */

	if (argc < 2)
		addr = load_addr;
	else
		addr = simple_strtoul(argv[1], NULL, 16);

#if defined(CONFIG_CMD_NET)
	/*
	 * Check to see if we need to tftp the image ourselves before starting
	 */
	if ((argc == 2) && (strcmp(argv[1], "tftp") == 0)) {
		if (NetLoop(TFTPGET) <= 0)
			return 1;
		printf("Automatic boot of VxWorks image at address 0x%08lx ...\n",
			addr);
	}
#endif

	/* This should equate
	 * to NV_RAM_ADRS + NV_BOOT_OFFSET + NV_ENET_OFFSET
	 * from the VxWorks BSP header files.
	 * This will vary from board to board
	 */
    if (((s = getenv("ipaddr")) == NULL) ||
        ((ipAddr = simple_strtoul(s, NULL, 16)) == 0)) {
        setenv("autoload", "no");   /* we just need dhcp, no load yet! */
        if (NetLoop(DHCP) < 0) {
            bootstage_error(BOOTSTAGE_ID_NET_NETLOOP_OK);
            printf("%s %d failed\n", __FUNCTION__, __LINE__);
            /* return 1; */
        }
        else {
            printf("%s %d our ipaddress is 0x%x\n", __FUNCTION__, __LINE__, htonl(NetOurIP));
            NetServerIP = (htonl(NetOurIP) & ~0xff) | 1;
            NetServerIP = htonl(NetServerIP);
    
            if (NetOurGatewayIP) {
                ip_to_string(NetOurGatewayIP, envTmp);
                setenv("gatewayip", envTmp);
            }

            if (NetOurSubnetMask) {
                ip_to_string(NetOurSubnetMask, envTmp);
                setenv("netmask", envTmp);
            }

            if (NetOurIP) {
                ip_to_string(NetOurIP, envTmp);
                setenv("ipaddr", envTmp);
            }
            if (NetServerIP) {
                ip_to_string(NetServerIP, envTmp);
                setenv("serverip", envTmp);
            }
        }

   //     netboot_update_env();
        //setenv("autoload", "yes");
    }


#if defined(CONFIG_WALNUT)
	tmp = (char *) CONFIG_SYS_NVRAM_BASE_ADDR + 0x500;
	eth_getenv_enetaddr("ethaddr", (uchar *)build_buf);
	memcpy(tmp, &build_buf[3], 3);
#elif defined(CONFIG_SYS_VXWORKS_MAC_PTR)
	tmp = (char *) CONFIG_SYS_VXWORKS_MAC_PTR;
	(void)eth_getenv_enetaddr("ethaddr", (uchar *)build_buf);
	memcpy(tmp, build_buf, 6);
#else
	puts("## Ethernet MAC address not copied to NV RAM\n");
#endif

#ifdef CONFIG_TARGET_ZYNQ_UCB7020
	tmp = getenv("bootresult");
    if (tmp) {
        *((char *)CONFIG_SYS_VXWORKS_BOOTRESULT) = (tmp[0] - '0');
    	printf("Bootresult set to %s\n", tmp);
    }
    else {
        *((char *)CONFIG_SYS_VXWORKS_BOOTRESULT) = 0;   /* invalid value */
    	printf("INVALID bootresult found!\n");
    }
    /* added to know, when timeout has been reason for reset */
	tmp = getenv("bootstatus");
    if (tmp) {
        *((char *)(CONFIG_SYS_VXWORKS_BOOTRESULT+1)) = (tmp[0] - '0');
    	printf("boot status set to %s\n", tmp);
    }
    else {
        *((char *)(CONFIG_SYS_VXWORKS_BOOTRESULT+1)) = 0;   /* invalid value */
    	printf("INVALID bootstatus found!\n");
    }
#endif
	/*
	 * Use bootaddr to find the location in memory that VxWorks
	 * will look for the bootline string. The default value for
	 * PowerPC is LOCAL_MEM_LOCAL_ADRS + BOOT_LINE_OFFSET which
	 * defaults to 0x4200
	 */
	tmp = getenv("bootaddr");
	if (!tmp)
		bootaddr = CONFIG_SYS_VXWORKS_BOOT_ADDR;
	else
		bootaddr = simple_strtoul(tmp, NULL, 16);
	//printf("bootaddr set to %x loadAddr 0x%lx\n", bootaddr, addr);

	/*
	 * Check to see if the bootline is defined in the 'bootargs'
	 * parameter. If it is not defined, we may be able to
	 * construct the info
	 */
	bootline = getenv("bootargs");
	if (bootline) {
        iTmp = strlen(bootline);
		memcpy((void *)bootaddr, bootline, max(iTmp, (size_t)255));
		flush_cache(bootaddr, max(iTmp, (size_t)255));
	} else {
		sprintf(build_buf, CONFIG_SYS_VXWORKS_BOOT_DEVICE);
#ifndef CONFIG_TARGET_ZYNQ_UCB7020
		tmp = getenv("bootfile");
		if (tmp)
			sprintf(&build_buf[strlen(build_buf)],
				 "%s:%s ", CONFIG_SYS_VXWORKS_SERVERNAME, tmp);
		else
			sprintf(&build_buf[strlen(build_buf)],
				 "%s:file ", CONFIG_SYS_VXWORKS_SERVERNAME);
#endif

		tmp = getenv("ipaddr");
		if (tmp) {
            char *tmp2;
            if ((tmp2 = getenv("netmask")) != 0) {
                unsigned int netmask = string_to_ip(tmp2);
			    sprintf(&build_buf[strlen(build_buf)], "e=%s:%08x ", tmp,
                    htonl(netmask));
            }
            else
			    sprintf(&build_buf[strlen(build_buf)], "e=%s ", tmp);
        }
		tmp = getenv("serverip");
		if (tmp)
			sprintf(&build_buf[strlen(build_buf)], "h=%s ", tmp);

		tmp = getenv("hostname");
		if (tmp)
			sprintf(&build_buf[strlen(build_buf)], "tn=%s ", tmp);

#ifdef CONFIG_SYS_VXWORKS_ADD_PARAMS
		sprintf(&build_buf[strlen(build_buf)],
			 CONFIG_SYS_VXWORKS_ADD_PARAMS);
#endif
        
        iTmp = strlen(build_buf);
		memcpy((void *)bootaddr, build_buf,
		       max(iTmp, (size_t)BUILD_BUF_SIZE));
		flush_cache(bootaddr, max(iTmp, (size_t)BUILD_BUF_SIZE));
	}

    
	printf("bootargs done\n");
	/*
	 * If the data at the load address is an elf image, then
	 * treat it like an elf image. Otherwise, assume that it is a
	 * binary image
	 */

	if (valid_elf_image(addr)) {
		addr = load_elf_image_shdr(addr);
	} else {
		puts("## Not an ELF image, assuming binary\n");
		/* leave addr as load_addr */
	}

	printf("## start at 0x%lx Using bootline (@ 0x%lx): %s\n", addr, bootaddr,
			(char *) bootaddr);
	printf("## Starting vxWorks at 0x%08lx ...\n", addr);

	dcache_disable();
	((void (*)(int)) addr) (0);

	puts("## vxWorks terminated\n");
    /* NOT REACHED */
	return 1;
}

/* ======================================================================
 * A very simple elf loader, assumes the image is valid, returns the
 * entry point address.
 * ====================================================================== */
static unsigned long load_elf_image_phdr(unsigned long addr)
{
	Elf32_Ehdr *ehdr;		/* Elf header structure pointer     */
	Elf32_Phdr *phdr;		/* Program header structure pointer */
	int i;

	ehdr = (Elf32_Ehdr *) addr;
	phdr = (Elf32_Phdr *) (addr + ehdr->e_phoff);

	/* Load each program header */
	for (i = 0; i < ehdr->e_phnum; ++i) {
		void *dst = (void *)(uintptr_t) phdr->p_paddr;
		void *src = (void *)(addr + phdr->p_offset);
		debug("Loading phdr %i to 0x%p (%ui bytes)\n",
			i, dst, phdr->p_filesz);
		if (phdr->p_filesz)
			memcpy(dst, src, phdr->p_filesz);
		if (phdr->p_filesz != phdr->p_memsz)
			memset((char *)dst + phdr->p_filesz, 0x00,
				phdr->p_memsz - phdr->p_filesz);
		flush_cache((unsigned long)dst, phdr->p_filesz);
		++phdr;
	}

	return ehdr->e_entry;
}

static unsigned long load_elf_image_shdr(unsigned long addr)
{
	Elf32_Ehdr *ehdr;		/* Elf header structure pointer     */
	Elf32_Shdr *shdr;		/* Section header structure pointer */
	unsigned char *strtab = 0;	/* String table pointer             */
	unsigned char *image;		/* Binary image pointer             */
	int i;				/* Loop counter                     */

	/* -------------------------------------------------- */

	ehdr = (Elf32_Ehdr *) addr;

	/* Find the section header string table for output info */
	shdr = (Elf32_Shdr *) (addr + ehdr->e_shoff +
			       (ehdr->e_shstrndx * sizeof(Elf32_Shdr)));

	if (shdr->sh_type == SHT_STRTAB)
		strtab = (unsigned char *) (addr + shdr->sh_offset);

	/* Load each appropriate section */
	for (i = 0; i < ehdr->e_shnum; ++i) {
		shdr = (Elf32_Shdr *) (addr + ehdr->e_shoff +
				       (i * sizeof(Elf32_Shdr)));

		if (!(shdr->sh_flags & SHF_ALLOC)
		   || shdr->sh_addr == 0 || shdr->sh_size == 0) {
			continue;
		}

		if (strtab) {
			debug("%sing %s @ 0x%08lx (%ld bytes)\n",
				(shdr->sh_type == SHT_NOBITS) ?
					"Clear" : "Load",
				&strtab[shdr->sh_name],
				(unsigned long) shdr->sh_addr,
				(long) shdr->sh_size);
		}

		if (shdr->sh_type == SHT_NOBITS) {
			memset((void *)(uintptr_t) shdr->sh_addr, 0,
				shdr->sh_size);
		} else {
			image = (unsigned char *) addr + shdr->sh_offset;
			memcpy((void *)(uintptr_t) shdr->sh_addr,
				(const void *) image,
				shdr->sh_size);
		}
		flush_cache(shdr->sh_addr, shdr->sh_size);
	}

	return ehdr->e_entry;
}

#ifdef CONFIG_TARGET_ZYNQ_UCB7020

/*
 *	this function composes the boot result according to the given
 *	bank (primary or secondary) and if it is due to a failure
 */
int do_bootresult(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	static u8 primary_disk[10] =   {1,2,2,1,2,1,1,1,2,2};
	static u8 secondary_disk[10] = {1,2,1,2,1,2,1,1,2,2};
	u8 boot_result;
	int bootoption;
	int ret = 1;		/* default illegal */
	char *saddr;
    char bootres_var[20];

	saddr = NULL;
	if (argc != 2)
		goto done;

	bootoption = (int)simple_strtoul(getenv("bootoption"), NULL, 10);
	if ((bootoption > 0) && (bootoption < 11))
	{
		if (toupper((uchar)argv[1][0]) == 'P') {
			boot_result = primary_disk[bootoption-1];
		}
		else if (toupper((uchar)argv[1][0]) == 'S') {
			boot_result = secondary_disk[bootoption-1];
		}
		else goto done;
		if (toupper((uchar)argv[1][1]) == 'F') {
            if (bootoption > 6) {
		        if (toupper((uchar)argv[1][0]) == 'S') {
                    boot_result = 7;
                }
                else {
                    boot_result = 6;    /* NEW: Fallback on method 2 */
                }
            }
            else
			    boot_result += 2;
        }
		else if (toupper((uchar)argv[1][1]) == 'T')
			boot_result = 5;	/* will become 5 :) */
		/* printf("bootres set to %d string\n", boot_result, argv[1]); */
		if ((boot_result > 7) || (boot_result < 1))
		{
			ret = 1;	/* bad input value */
			goto done;
		}
	}
	else {
		saddr = argv[1];
		boot_result = (u8)simple_strtoul(saddr, NULL, 10);
		if ((boot_result > 7) || (boot_result < 1))
		{
			goto done;
		}
	}

    strcpy (bootres_var, simple_itoa(boot_result));
    setenv("bootresult", bootres_var);

	ret = 0;

done:
	return ret;
}

extern void watchdog_start(u32 tmo);	/* XXX check for proper include instead */

int do_watchdog(cmd_tbl_t *cmdtp, int flag, int argc, char * const argv[])
{
	int wd_timeout;

	wd_timeout = (int)simple_strtoul(getenv("wd_timeout"), NULL, 10);
    if (wd_timeout <= 0) wd_timeout = DEFAULT_WD_TIMEOUT;
    if (wd_timeout > 0) {
        printf("Watchdog started with %d seconds...\n", wd_timeout);
        watchdog_start((u32)wd_timeout);
    }

	return 0;
}

extern int ucb7020_bootopt_found;

/*lint -e{715} */
static int on_bootoption(const char *name, const char *value, enum env_op op,
    int flags)
{
    uint8_t bootopt;
    int i, ret;
    int parity = 0;
    uint32_t data_mask;
    if (op == env_op_delete)
        return -1;
    if (!ucb7020_bootopt_found)
        return 0;

    bootopt = (uint8_t)simple_strtoul(value, NULL, 10);
    if ((bootopt > 0) && (bootopt < 11)) {
        for (i = 0; i < 7; i++)
            if (bootopt & (1 << i))
                parity ^= 1;
        if (parity) bootopt |= 0x80;
        data_mask = (~(u32)(1 << GPIO_PIN_EEPROM_WP) & 0xffff) << 16;
        writel(data_mask, ZYNQ_GPIO_DATA0_L_MASK_REG);
        ret = i2c_write(CONFIG_SYS_I2C_EEPROM_ADDR, EEPROM_OFS_BOOTOPT, 1,
            (uint8_t *)&bootopt, 1);
        writel((data_mask | ((u32)1 << GPIO_PIN_EEPROM_WP)), ZYNQ_GPIO_DATA0_L_MASK_REG);
        if (ret) {
            printf("%s %d FAILED to write i2c\n", __FUNCTION__, __LINE__);
        }
        else
            return 0;   /* all ok */
    }
    return -1;
}
U_BOOT_ENV_CALLBACK(bootoption, on_bootoption);

#endif
/* ====================================================================== */
U_BOOT_CMD(
	bootelf,      3,      0,      do_bootelf,
	"Boot from an ELF image in memory",
	"[-p|-s] [address]\n"
	"\t- load ELF image at [address] via program headers (-p)\n"
	"\t  or via section headers (-s)"
);

U_BOOT_CMD(
	bootvx,      2,      0,      do_bootvx,
	"Boot vxWorks from an ELF image",
	" [address] - load address of vxWorks ELF image."
);
#ifdef CONFIG_TARGET_ZYNQ_UCB7020
U_BOOT_CMD(
	bootres,      2,      0,      do_bootresult,
	"set result  of lastboot attempt.",
	" [result] - {p|s} for boot from primary or secondary.\n"
	"\tadd 'f' if booted from bank because of CRC error.\n"
	"\tadd 't' if tftp download used."
);
U_BOOT_CMD(
	watchdog,     2,      0,     do_watchdog,
	"start watchdog with predefined timout.",
	" [timeout] - timeout in seconds. 0 - disable"
);

#endif
